"""
Database module
"""


class Database:
    """
    Defines data structures and methods to store article content.
    """

    def save(self, article):
        """
        Saves an article.

        Args:
            article: article metadata and text content
        """

    def complete(self):
        """
        Signals processing is complete and runs final storage methods.
        """

    def close(self):
        """
        Commits and closes the database.
        """
